package net.osdn.util.ssdp.client;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.ConnectException;
import java.net.DatagramPacket;
import java.net.Inet4Address;
import java.net.Inet6Address;
import java.net.InetAddress;
import java.net.InterfaceAddress;
import java.net.MulticastSocket;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.net.SocketTimeoutException;
import java.net.URL;
import java.net.URLConnection;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;

import net.osdn.util.ssdp.Device;

/** SSDPプロトコルを使ってネットワーク上のデバイスを発見します。
 * 
 */
public class SsdpClient {
	/** デバッグ・フラグ。trueを設定すると詳細ログが標準出力および標準エラー出力に出力されるようになります。
	 */
	public static boolean DEBUG = false;

	/** SSDP M-SEARCHリクエストをマルチキャスト送信する宛先ポート番号です。
	 */
	public static final int SSDP_PORT = 1900;
	
	private static final Charset UTF_8 = Charset.forName("UTF-8");
	private static final int RECV_TIMEOUT_MIN = 5000;
	private static final int RECV_TIMEOUT = 15000;
	private static final int SEND_INTERVAL = 3000;
	
	private int timeout;
	private DiscoveryCallback callback;
	private MulticastSocket socket;
	private Thread sendThread;
	private Thread recvThread;
	private Set<Device> foundDevices = new HashSet<Device>();
	private List<Future<Device>> futures = new ArrayList<Future<Device>>();
	private boolean isSendRunning;
	private boolean isRecvRunning;
	private boolean isCancelRequested;
	private List<Device> list;
	
	public SsdpClient() {
	}

	/** 探索(discover)をキャンセルします。
	 * このメソッドを呼び出すと getメソッドの待機状態から復帰します。
	 * 
	 */
	public void cancel() {
		if(!isCancelRequested && (isSendRunning || isRecvRunning)) {
			isCancelRequested = true;

			if(socket != null) {
				try {
					socket.close();
				} catch(Exception e) {
					if(DEBUG) {
						e.printStackTrace();
					}
				} finally {
					socket = null;
				}
			}

			try {
				sendThread.interrupt();
			} catch(Exception e) {
				if(DEBUG) {
					e.printStackTrace();
				}
			}
			try {
				recvThread.interrupt();
			} catch(Exception e) {
				if(DEBUG) {
					e.printStackTrace();
				}
			}
		}
	}
	
	/** 探索(discover)が指定時間が経過して完了するか、もしくはキャンセルされるまで待機して、見つかったデバイスのリストを返します。
	 * 
	 * @return 見つかったデバイスのリスト
	 */
	public List<Device> get() {
		
		if(list != null) {
			return list;
		}
		
		if(isSendRunning) {
			try {
				sendThread.join();
			} catch (InterruptedException e) {
				if(DEBUG) {
					e.printStackTrace();
				}
			} finally {
				sendThread = null;
			}
		}
		
		if(isRecvRunning) {
			try {
				recvThread.join();
			} catch (InterruptedException e) {
				if(DEBUG) {
					e.printStackTrace();
				}
			} finally {
				recvThread = null;
			}
			
		}
			
		if(socket != null) {
			try {
				socket.close();
			} catch(Exception e) {
				if(DEBUG) {
					e.printStackTrace();
				}
			} finally {
				socket = null;
			}
		}
			
		if(futures != null) {
			list = new ArrayList<Device>();
			for(Future<Device> future : futures) {
				Device device = null;
				try {
					device = future.get();
				} catch (InterruptedException e) {
					if(DEBUG) {
						e.printStackTrace();
					}
				} catch (ExecutionException e) {
					Throwable cause = e.getCause();
					if(cause instanceof ConnectException) {
						// System.err.println(cause.getClass().getName() + ": " + cause.getMessage());
					} else {
						if(DEBUG) {
							e.printStackTrace();
						}
					}
				}
				
				if(device != null) {
					list.add(device);
				}
			}
		}
		
		return list;
	}
	
	/** 既定の時間(15秒)でデバイス検索を開始します。
	 * デバイス検索中は3秒ごとに繰り返しM-SEARCHリクエストがマルチキャスト送信されます。
	 * 検索結果を取得するために getメソッドを呼び出す必要があります。
	 * 
	 * @throws IOException 例外が発生した場合
	 */
	public void discover() throws IOException {
		discover(RECV_TIMEOUT, null);
	}
	
	/** 指定した時間でデバイス検索を開始します。
	 * デバイス検索中は3秒ごとに繰り返しM-SEARCHリクエストがマルチキャスト送信されます。
	 * 検索結果を取得するために getメソッドを呼び出す必要があります。
	 * 
	 * @param timeout 検索時間(ミリ秒)
	 * @throws IOException 例外が発生した場合
	 */
	public void discover(int timeout) throws IOException {
		discover(timeout, null);
	}
	
	/** コールバック・インターフェースを指定して既定の時間(15秒間)デバイスを検索します。
	 * デバイスからの応答があるたびにコールバック・インターフェースのdicoveredメソッドが呼び出され、見つかったデバイス情報を逐次受け取ることができます。
	 * デバイス検索中は3秒ごとに繰り返しM-SEARCHリクエストがマルチキャスト送信されます。
	 * 検索結果を取得するために getメソッドを呼び出すこともできます。(任意操作)
	 * 
	 * @param callback デバイス検索の結果を都度受け取るためのコールバック・インターフェース
	 * @throws IOException 例外が発生した場合
	 */
	public void discover(DiscoveryCallback callback) throws IOException {
		discover(RECV_TIMEOUT, callback);
	}

	/** コールバック・インターフェースと検索時間を指定してデバイスを検索します。
	 * デバイスからの応答があるたびにコールバック・インターフェースのdicoveredメソッドが呼び出され、見つかったデバイス情報を逐次受け取ることができます。
	 * デバイス検索中は3秒ごとに繰り返しM-SEARCHリクエストがマルチキャスト送信されます。
	 * 検索結果を取得するために getメソッドを呼び出すこともできます。(任意操作)
	 * 
	 * @param timeout 検索時間(ミリ秒)
	 * @param callback デバイス検索の結果を都度受け取るためのコールバック・インターフェース
	 * @throws IOException 例外が発生した場合
	 */
	public void discover(int timeout, DiscoveryCallback callback) throws IOException {
		if(isSendRunning || isRecvRunning) {
			return;
		}
		isCancelRequested = false;
		
		this.timeout = (timeout > RECV_TIMEOUT_MIN) ? timeout : RECV_TIMEOUT_MIN;
		this.callback = callback;
		this.list = null;
		this.foundDevices.clear();
		
		socket = new  MulticastSocket(0);
			
		sendThread = new Thread() {
			@Override
			public void run() {
				try {
					send();
				} catch(Exception e) {
					if(!isCancelRequested) {
						throw new RuntimeException(e);
					}
				}
				isSendRunning = false;
			}
		};
		sendThread.start();
		isSendRunning = true;
			
		recvThread = new Thread() {
			@Override
			public void run() {
				try {
					recv();
				} catch(Exception e) {
					if(!isCancelRequested) {
						throw new RuntimeException(e);
					}
				}
				isRecvRunning = false;
				if(SsdpClient.this.callback != null) {
					try {
						SsdpClient.this.callback.done();
					} catch(Exception e) {
						if(DEBUG) {
							e.printStackTrace();
						}
					}
				}
			};
		};
		recvThread.start();
		isRecvRunning = true;
	}
	
	/** マルチキャストをサポートしているすべてのネットワークインターフェースに対して、
	 * 3秒ごとに繰り返しM-SEARCHリクエストをマルチキャスト送信します。
	 * 
	 * @throws InterruptedException 割り込みが発生した場合
	 */
	protected void send() throws InterruptedException {
		String strIpv4mcastaddr = "239.255.255.250";
		String strIpv6mcastaddr = "FF02::C";
		String query = "M-SEARCH * HTTP/1.1\r\n"
						   + "Host:${HOST}:${PORT}\r\n"
						   + "ST:upnp:rootdevice\r\n"
						   + "MAN:\"ssdp:discover\"\r\n"
						   + "MX:1\r\n"
						   + "\r\n";
		
		int count = (timeout - 4000) / SEND_INTERVAL;
		if(count < 1) {
			count = 1;
		}
		
		InetAddress ipv4mcastaddr = null;
		DatagramPacket ipv4packet = null;
		try {
			byte[] buf = query.replace("${HOST}", strIpv4mcastaddr).replace("${PORT}", Integer.toString(SSDP_PORT)).getBytes();
			ipv4mcastaddr = InetAddress.getByName(strIpv4mcastaddr);
			ipv4packet = new DatagramPacket(buf, 0, buf.length, ipv4mcastaddr, SSDP_PORT);
		} catch(Exception e) {
			if(DEBUG) {
				e.printStackTrace();
			}
		}
		
		InetAddress ipv6mcastaddr = null;
		DatagramPacket ipv6packet = null;
		try {
			byte[] buf = query.replace("${HOST}", "[" + strIpv6mcastaddr + "]").replace("${PORT}", Integer.toString(SSDP_PORT)).getBytes();
			ipv6mcastaddr = InetAddress.getByName(strIpv6mcastaddr);
			ipv6packet = new DatagramPacket(buf, 0, buf.length, ipv6mcastaddr, SSDP_PORT);
		} catch(Exception e) {
			if(DEBUG) {
				e.printStackTrace();
			}
		}

		for(int i = 0; i < count; i++) {
			try {
				Enumeration<NetworkInterface> interfaces = NetworkInterface.getNetworkInterfaces();
				while(interfaces.hasMoreElements()) {
					NetworkInterface netIf = interfaces.nextElement();
					if(netIf.isLoopback() || netIf.isVirtual() || !netIf.isUp() || !netIf.supportsMulticast()) {
						continue;
					}
					for(InterfaceAddress ifAddr : netIf.getInterfaceAddresses()) {
						InetAddress addr = ifAddr != null ? ifAddr.getAddress() : null;
						if(isCancelRequested) {
							return;
						}
						if(ipv4mcastaddr != null && addr instanceof Inet4Address) {
							if(DEBUG) {
								System.out.println("# SSDP: send: IPv4: netIf=" + netIf + ", ifAddr=" + ifAddr + ", addr=" + addr + ", isLoopback=" + netIf.isLoopback() + ", isVirtual=" + netIf.isVirtual() + ", netIf.isUp=" + netIf.isUp() + ", supportMulticast=" + netIf.supportsMulticast());
							}
							try {
								socket.setInterface(addr);
								socket.send(ipv4packet);
							} catch(Exception e) {
								if(!isCancelRequested) {
									System.err.println("# SSDP: send: IPv4: ERROR, ifAddr=" + ifAddr);
									e.printStackTrace();
								}
							}
						}
						if(ipv6mcastaddr != null && addr instanceof Inet6Address) {
							if(DEBUG) {
								System.out.println("# SSDP: send: IPv6: netIf=" + netIf + ", ifAddr=" + ifAddr + ", addr=" + addr + ", isLoopback=" + netIf.isLoopback() + ", isVirtual=" + netIf.isVirtual() + ", netIf.isUp=" + netIf.isUp() + ", supportMulticast=" + netIf.supportsMulticast());
							}
							try {
								socket.setInterface(addr);
								socket.send(ipv6packet);
							} catch(Exception e) {
								if(!isCancelRequested) {
									System.err.println("# SSDP: send: IPv6: ERROR, ifAddr=" + ifAddr);
									e.printStackTrace();
								}
							}
						}
					}
				}
				if(i + 1 < count) {
					Thread.sleep(SEND_INTERVAL);
				}
			} catch(IOException e) {
				e.printStackTrace();
			}
		}
	}
	
	/** M-SEARCHリクエストに対する応答パケットを受信します。
	 * 応答パケットにLocationヘッダーが含まれる場合、そのLocationにHTTP GETリクエストを送信してデバイス情報を取得します。
	 * 取得されたデバイス情報は、コールバック・インターフェースを通して通知されます。(設定されている場合)
	 * 取得されたデバイス情報はリストとして保持され、getメソッド呼び出しでまとめて取得することができます。
	 * 
	 * @throws InterruptedException 割り込みが発生した場合
	 * @throws IOException 例外が発生した場合
	 */
	protected void recv() throws InterruptedException, IOException {
		ExecutorService executor = null;
		try {
			executor = Executors.newCachedThreadPool();
			futures.clear();
			
			List<InetAddress> localAddresses = null;
			try {
				localAddresses = getLocalAddresses();
			} catch(SocketException e) {
				e.printStackTrace();
			}
			
			long startTime = System.currentTimeMillis();
			byte[] buf = new byte[1024];
			DatagramPacket packet = new DatagramPacket(buf, buf.length);
			for(;;) {
				try {
					int d = (int)(System.currentTimeMillis() - startTime);
					if(timeout - d <= 0) {
						break;
					}
					socket.setSoTimeout(timeout - d);
					socket.receive(packet);
				} catch(SocketTimeoutException e) {
					break;
				}
				if(packet.getLength() > 0 && packet.getData()[0] == 'H') {
					SearchResponse response = new SearchResponse(packet);
					if(response.getStatus() == 200) {
						final InetAddress deviceAddr = packet.getAddress();
						final String location = response.getHeader("Location");
						final DiscoveryCallback dc = callback;
						final boolean isSelf = (localAddresses != null && localAddresses.contains(deviceAddr));
						Future<Device> future = executor.submit(new Callable<Device>() {
							@Override
							public Device call() throws Exception {
								URL url = new URL(location);
								URLConnection connection = null;
								InputStream in = null;
								BufferedReader r = null;
								try {
									connection = url.openConnection();
									in = connection.getInputStream();
									r = new BufferedReader(new InputStreamReader(in, UTF_8));
									Device device = Device.parse(r);
									device.setAddress(deviceAddr);
									device.setSelf(isSelf);
									synchronized(foundDevices) {
										if(foundDevices.contains(device)) {
											return null;
										}
										foundDevices.add(device);
									}
									if(dc != null) {
										try {
											dc.discovered(device);
										} catch(Exception e) {
											if(DEBUG) {
												e.printStackTrace();
											}
										}
									}
									return device;
								} finally {
									if(r != null) {
										try { r.close(); } catch(Exception e) {}
									}
									if(in != null) {
										try { in.close(); } catch(Exception e) {}
									}
								}
							}
						});
						futures.add(future);
					}
				}
			}
		} finally {
			if(executor != null) {
				executor.shutdown();
			}
		}
	}
	
	protected void shutdown() {
		
	}
	
	
	/** ローカルアドレスのリストを取得します。
	 * 
	 * @return
	 * @throws SocketException
	 */
	private static List<InetAddress> getLocalAddresses() throws SocketException {
		List<InetAddress> addresses = new ArrayList<InetAddress>();
		
		Enumeration<NetworkInterface> interfaces = NetworkInterface.getNetworkInterfaces();
		while(interfaces.hasMoreElements()) {
			NetworkInterface network = interfaces.nextElement();
			for(InterfaceAddress ifAddr : network.getInterfaceAddresses()) {
				InetAddress addr = ifAddr.getAddress();
				if(addr != null) {
					addresses.add(addr);
				}
			}
		}
		
		return addresses;
	}
}
